#ifndef _PIN_HPP
#define _PIN_HPP

#include <avr/io.h>

/** expands PINDEF(B, 0) to &DDRB, &PORTB, &PINB, PB0
 * example: 
 *      OutputPin ledPin ( PINDEF(B, 0) );
 *   expands to: 
 *      OutputPin ledPin (&DDRB, &PORTB, &PINB, PB0);
 */
#define PINDEF(X,Y)  &DDR##X,&PORT##X,&PIN##X,Y


typedef volatile uint8_t* _Register;


class Pin { /** base class for a pin */
public:
    _Register ddr;
    _Register port;
    _Register pin;
    uint8_t  num;

    /** constructor 
     * @param  ddr   DDRn register address of pin (eg. &DDRB)
     * @param  port  PORTn register ''            (eg. &PORTB)
     * @param  pin   PINn register                (eg. &PINB)
     * @param  num   pin number                   (eg. PB2)
     */
    Pin(_Register ddr, _Register port, _Register pin, uint8_t num):
	ddr(ddr), port(port), pin(pin), num(num)
    {}
};


class InputPin : public Pin { /** an input pin */
public:
    /** constructor
     * @params  see Pin::Pin()
     */
    InputPin(_Register ddr, _Register port, 
	     _Register pin, uint8_t num):
        Pin(ddr, port, pin, num) {
        *this->ddr &= ~(1 << this->num);
    }

    /** return true if pin is high, else false */
    bool isHigh(void) { return (*this->pin & (1 << this->num)); };

    /** return true if pin is low, else false */
    bool isLow(void) { return !((*this->pin & (1 << this->num))); };
};


class OutputPin : public Pin { /** an output pin */
public:
    OutputPin(_Register ddr, _Register port, 
	      _Register pin, uint8_t num):
	/** constructor @params  see Pin::Pin() */
        Pin(ddr, port, pin, num) {
        *this->ddr |= (1 << this->num);
    }

    OutputPin& setHigh(void) { /** set pin high */
	*this->port |= (1 << this->num); return *this; };

    OutputPin& setLow(void)  { /** set pin low */
	*this->port &= ~(1 << this->num); return *this; };

    /** assignment operator 
     * @param  parm  if true: set pin high, else set pin low
     */
    void operator=(bool const &parm) {
        if(parm) {
            this->setHigh();
        } else {
            this->setLow();
        }
    }
};


/**
 * deriving this from InputPin and OutputPin by virtually
 * inheriting from Pin bloats code so we rather duplicate
 */
class InoutPin: public Pin {
public:
    /** constructor
     * @params  see Pin::Pin()
     */
    InoutPin(_Register ddr, _Register port, _Register pin, uint8_t num):
	Pin(ddr, port, pin, num)
    {}

    /** set pin as input pin */
    InoutPin& setInput(void) { *this->ddr &= ~(1 << this->num); return *this; }

    /** set pin as output pin */
    InoutPin& setOutput(void) { *this->ddr |= (1 << this->num); return *this; }

    // input functions (see InputPin)
    int isHigh(void) { return (*this->pin & (1 << this->num)); };
    int isLow(void) { return !((*this->pin & (1 << this->num))); };

    // output functions (see OutputPin)
    InoutPin setHigh(void) { *this->port |= (1 << this->num); return *this; };
    InoutPin setLow(void)  { *this->port &= ~(1 << this->num); return *this; };
    void operator=(bool const &parm) {
        if(parm) {
            this->setHigh();
        } else {
            this->setLow();
        }
    }
};

#endif // _PIN_HPP
